"""Module pdfsyntax.display: pretty print the PDF file structure as HTML"""

import html
from .objects import Stream


NAME_MAX_WIDTH = 20
VALUE_MAX_WIDTH = 30

HEADER = '''
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>pdfsyntax</title>
    <style>
        body {
            font-family: monospace;
        }
        :target {
            background-color: LightYellow;
        }
        .obj-header {
            background-color: LightGray;
        }
        .obj-body {
            background-color: WhiteSmoke;
            margin: 0 0 0 0;
        }
        .obj-low {
            color: Gray;
        }
        .obj-link {
            background-color: AliceBlue;
        }
        .eof {
            background-color: AntiqueWhite;
        }
        .header-button {
            background-color: AliceBlue;
            color: Grey;
        }
        .header-link {
            color: Grey;
        }
        .important {
            background-color: AntiqueWhite;
        }
        .header {
            position: fixed;
            top: 0px;
            left: 0px;
            width: 100%;
            background-color: Black;
            color: White;
        }
        .content {
            margin-top: 5em;
        }
        pre {
            margin-top: 0.5em;
            margin-bottom: 0.5em;
        }
    </style>
</head>
<body>
'''

TRAILER = '''
</div>
</body>
'''

TRUNCATED = '<em> ...(truncated) </em>'


def pos2ref_from_index(index: list, abs_pos: int) -> dict:
    """."""
    for _, current in enumerate(index):
        for x in current:
            if x is None:
                continue
            if abs_pos == x.get('abs_pos') and x.get('o_num') != 0:
                o_num = x['o_num']
                o_gen = x['o_gen']
                o_ver = x['o_ver']
                return (o_num, o_gen, o_ver)
    return None


def recent_ref_from_index(index: list, iref: complex) -> dict:
    """."""
    res = None
    o_num = int(iref.imag)
    for _, current in enumerate(index):
        x = current[o_num]
        if x is None:
            continue
        else:
            res =  (o_num, x['o_gen'], x['o_ver'])
    return res


def build_html(articles: list, index: list, filename: str, version: str) -> str:
    """Compose the page layout."""
    nb_ver = len(index)
    last = articles[-1][0]
    page = HEADER
    page += build_header(filename, last, version)
    for article in articles:
        pos, _, typ, content = article
        if typ == 'STARTXREF':
            page += add_startxref(article, index)
        elif typ == 'COMMENT' and content == b'%%EOF':
            page += add_eof(article)
        elif typ == 'IND_OBJ':
            page += build_obj_header(article, index)
            page += follow_obj(content['obj'], index)
            page += build_obj_trailer()
        elif typ == 'XREFTABLE':
            page += build_obj_header(article, index)
            page += build_xref_table(content['table'], index)
            page += "\ntrailer\n"
            page += follow_obj(content['trailer'], index)
            page += build_obj_trailer()
        elif typ == 'XREF' and content[0] == 'XREF_S':
            page += build_xref_item_header()
            page += build_xref_stream_item(content, index)
            page += build_obj_trailer()
    page += TRAILER
    return page


def build_header(filename: str, last: int, version: bytes) -> str:
    """Add a banner with the file name."""
    ret = ''
    ret += f'<div class="header">\n'
    ret += f'<pre> <a class="header-button" href="#abs{last}">jump to end</a>'
    ret += f'     <span class="obj-low">Hypertext inspection of </span>{filename}<span class="obj-low"> generated by </span>'
    ret += f'<a class="header-link" href="https://pdfsyntax.dev">pdfsyntax</a></pre>'
    ret += f'</div>\n'
    ret += f'<div class="content">\n'
    ret += f'<pre>{version[:8].decode("ascii")}</pre>\n'
    return ret


def add_startxref(article: dict, index: list) -> str:
    """ """
    pos, _, _, xref = article
    ref = pos2ref_from_index(index, int(xref))
    if ref is None:
        href = f'abs{int(xref)}'
    else:
        o_num, o_gen, o_ver = ref
        href = f'obj{o_num}.{o_gen}.{o_ver}'
    ret = ''
    ret += f'<div id="abs{pos}">\n<pre>\n\n\n'
    ret += f'startxref\n'
    if xref == 0:
        ret += f'0\n'
    else:
        ret += f'<a class="obj-link" href="#{href}">{int(xref)}</a>\n'
    ret += f'\n</pre>\n</div>\n'
    return ret


def add_eof(article: dict) -> str:
    """ """
    pos, _, _, _ = article
    ret = ''
    ret += f'<div id="abs{pos}">\n<pre class="eof">\n'
    ret += f'%%EOF\n'
    ret += f'</pre>\n</div>\n'
    return ret


def build_xref_table(table: list, index: list) -> str:
    """Display XREF table with additional links to objects."""
    ret = ''
    for x in table:
        if len(x) == 2:
            start, size = x
            ret += f'{start}  {size}'
        elif len(x) == 4:
            pos, o_num, o_gen, st = x
            if st == b'f':
                continue
            _, _, o_ver = recent_ref_from_index(index, complex(o_gen, o_num))
            ret += f'{pos:010} {o_gen:05} {st.decode("ascii")}'
            if o_num != 0:
                ret += '    '
                ret += f'<a href="#obj{o_num}.{o_gen}.{o_ver}">'
                ret += f'<span class="obj-link">#{o_num} {o_gen}</span>'
                ret += '</a>'
        ret += '\n'
    return ret


def build_xref_stream_item(item: tuple, index: list) -> str:
    """Display XREF stream item."""
    ret = ' '
    _, pos, o_num, o_gen, st, env_num, raw_line = item
    _, _, o_ver = recent_ref_from_index(index, complex(o_gen, o_num))
    if o_num != 0:
        ret += f'<a href="#obj{o_num}.{o_gen}.{o_ver}">'
        ret += f'<span class="obj-link">#{o_num} {o_gen}</span>'
        ret += '</a>'
        ret += '    '
        if env_num:
            ret += f'In object stream {env_num} at {pos:010} {st.decode("ascii")}'
        else:
            ret += f'At absolute position {pos:010} {st.decode("ascii")}'
    ret += '\n'
    return ret


def build_xref_stream(table: list, mini_index: list) -> str:
    """Display XREF stream with additional links to objects."""
    ret = '\nstream\n'
    for line, o_num in table:
        ret += line.decode('ascii')
        if o_num != None:
            o_gen, o_ver = mini_index[o_num]
            ret += '    '
            ret += f'<a href="#obj{o_num}.{o_gen}.{o_ver}">'
            ret += f'<span class="obj-link">#{o_num} {o_gen}</span>'
            ret += '</a>'
        ret += '\n'
    return ret


def move_list_item(mod_list: list, item: int, new_pos: int) -> str:
    """Reposition an item in a list."""
    if item in mod_list:
        old_pos = mod_list.index(item)
        mod_list.pop(old_pos)
        mod_list.insert(new_pos, item)
    return mod_list


def follow_obj(obj, index: list, depth=0) -> str:
    """Recursively construct object representation."""
    ret = ''
    content = None
    if isinstance(obj, complex):
        o_num, o_gen, o_ver = recent_ref_from_index(index, obj)
        ret += f'<a href="#obj{o_num}.{o_gen}.{o_ver}">'
        ret += f'<span class="obj-link">{o_num} {o_gen} R</span>'
        ret += '</a>'
        return ret
    if type(obj) == dict or type(obj) == Stream: 
        if type(obj) == Stream:
            content = obj['stream']
            obj = obj['entries']
        ret += '<<\n'
        keys = list(obj.keys())
        keys = move_list_item(keys, '/Type', 0)
        keys = move_list_item(keys, '/Subtype', 1)
        for i in keys:
            name = i #name = i.decode('ascii')
            value = follow_obj(obj[i], index, depth + 1)
            ret += ' ' * (NAME_MAX_WIDTH + 2) * depth
            if name == '/Type' or name == '/Subtype':
                ret += f'  {name:{NAME_MAX_WIDTH}}<span class="important">{value}</span>\n'
            elif name == '/Prev':
                ret += f'  {name:{NAME_MAX_WIDTH}}<a class="obj-link" href="#abs{value}">{value}</a>\n'
            else:
                ret += f'  {name:{NAME_MAX_WIDTH}}{value}\n'
        ret += ' ' * (NAME_MAX_WIDTH + 2) * depth
        ret += '>>'
        if content:
            content = f'{content}'[2:-1]
            content = content[:VALUE_MAX_WIDTH * 2] + TRUNCATED
            ret += f'  \n{"stream":{NAME_MAX_WIDTH}}\n{content}\n'
    elif type(obj) == list:
        ret += '[ '
        for i in obj:
            value = follow_obj(i, index)
            ret += f'{value} '
        if len(ret) > VALUE_MAX_WIDTH and 'href' not in ret:
            ret = ret[:VALUE_MAX_WIDTH] + TRUNCATED
        ret += ']'        
    else:
        comment = False
        if type(obj) == int or type(obj) == float or type(obj) == bool or type(obj) == str or obj is None:
            ret += str(obj)
        elif isinstance(obj, complex) == True:
            ret += str(obj.num)
        elif obj[0:1] == b'(':
            if len(obj) > VALUE_MAX_WIDTH:
                obj = obj[:VALUE_MAX_WIDTH]
                comment = True
            obj = f'{obj}'
            ret += f'{obj[2:-1]}'
        else:
            ret += obj.decode('ascii')
        ret = html.escape(ret)
        if comment == True:
            ret += TRUNCATED
    return ret


def build_xref_item_header() -> str:
    """."""
    ret = ''
    ret += f'<div><pre>'
    #ret += f'<div class="obj-body">\n'
    return ret


def build_obj_header(article, index) -> str:
    """Add opening elements to object."""
    pos, _, typ, obj = article
    ref = pos2ref_from_index(index, pos)
    ret = ''
    if ref:
        o_num, o_gen, o_ver = ref
        ret += f'<div id="obj{o_num}.{o_gen}.{o_ver}">\n<pre>\n\n\n'
        ret += f'<span class="obj-header"><strong>{o_num}</strong> <span class="obj-low">{o_gen} obj</span></span>'
        ret += f'<em class="obj-low">  at offset {pos}</em>'
    elif type(pos) == tuple:
        o_num, o_gen, o_ver = obj['o_num'], 0, 0
        ret += f'<div id="obj{o_num}.{o_gen}.{o_ver}">\n<pre>\n\n\n'
        ret += f'<span class="obj-header"><strong>{o_num}</strong> <span class="obj-low">{o_gen} obj</span></span>'
        ret += f'<em class="obj-low">  from object stream {obj["env_num"]} above</em>'
    else:
        ret += f'<div id="abs{pos}">\n<pre>\n\n\n'
        ret += f'<span class="obj-header"><strong>XREF table & trailer</strong></span>'
        ret += f'<em class="obj-low">  at offset {pos}</em>'
    ret += f'<div class="obj-body">\n'
    return ret


def build_obj_trailer() -> str:
    """Add closing elements to object."""
    ret = ''
    ret += f'</div>\n'
    ret += f'</pre>\n</div>\n'
    return ret



